# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.upgrade.rules.base_rule import BaseRule
from airflow.configuration import conf

invalid_config_keys = {
    "airflow_configmap",
    "airflow_local_settings_configmap",
    "dags_in_image",
    "dags_volume_subpath",
    "dags_volume_mount_point", "dags_volume_claim",
    "logs_volume_subpath", "logs_volume_claim",
    "dags_volume_host", "logs_volume_host",
    "env_from_configmap_ref", "env_from_secret_ref", "git_repo",
    "git_branch", "git_sync_depth", "git_subpath",
    "git_sync_rev", "git_user", "git_password",
    "git_sync_root", "git_sync_dest",
    "git_dags_folder_mount_point", "git_ssh_key_secret_name",
    "git_ssh_known_hosts_configmap_name", "git_sync_credentials_secret",
    "git_sync_container_repository",
    "git_sync_container_tag", "git_sync_init_container_name",
    "git_sync_run_as_user",
    "worker_service_account_name", "image_pull_secrets",
    "gcp_service_account_keys", "affinity",
    "tolerations", "run_as_user", "fs_group"
}


class PodTemplateFileRule(BaseRule):
    title = "Users must set a kubernetes.pod_template_file value"

    description = """\
In Airflow 2.0, KubernetesExecutor Users need to set a pod_template_file as a base
value for all pods launched by the KubernetesExecutor. Many Kubernetes configs are no longer
needed once this pod_template_file has been generated.
"""

    def should_skip(self):
        # Check this rule only if users use KubernetesExecutor
        if conf.get("core", "executor") != "KubernetesExecutor":
            return "Skipped because this rule applies only to environment using KubernetesExecutor."

    def check(self):
        pod_template_file = conf.get("kubernetes", "pod_template_file", fallback=None)
        if not pod_template_file:
            return (
                "Please create a pod_template_file by running `airflow generate_pod_template`.\n"
                "This will generate a pod using your aiflow.cfg settings"
            )

        conf_dict = conf.as_dict(display_sensitive=True)
        kube_conf = conf_dict['kubernetes']
        keys = kube_conf.keys()
        resp = [k for k in keys if k in invalid_config_keys]
        if conf_dict['kubernetes_labels']:
            resp.append("kubernetes_labels")
        if conf_dict['kubernetes_secrets']:
            resp.append("kubernetes_secrets")

        if resp:
            resp_string = "\n".join(resp)
            return "The following invalid keys were found in your airflow.cfg: \
                   \n\n{resp_string}\n\n \
                   Now that you have a pod_template_file, these keys no longer do anything.\n\
                   Please delete these keys.".format(resp_string=resp_string)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
import yaml
from jsonschema import validate

from airflow.utils.module_loading import import_string

SCHEMA = {
    "$schema": "http://json-schema.org/draft-04/schema#",
    "type": "object",
    "properties": {
        "ignored_rules": {"type": ["array", "null"], "items": {"type": "string"}},
        "custom_rules": {"type": ["array", "null"], "items": {"type": "string"}},
    },
    "additionalProperties": False,
}


class UpgradeConfig(object):
    def __init__(self, raw_config):
        self._raw_config = raw_config

    def get_ignored_rules(self):
        return self._raw_config.get("ignored_rules") or []

    def get_custom_rules(self):
        custom_rules = self._raw_config.get("custom_rules") or []
        return [import_string(r)() for r in custom_rules]

    @classmethod
    def read(cls, path):
        with open(path) as f:
            raw_config = yaml.safe_load(f)
        validate(raw_config, schema=SCHEMA)
        return cls(raw_config)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from sqlalchemy import func
from airflow.models import Connection
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.db import provide_session


class UniqueConnIdRule(BaseRule):
    title = "Connection.conn_id is not unique"

    description = """\
The `id` column in the `connection` table must be unique. Previously, this rule was \
enforced by application logic, but was not enforced by the database schema.

If you made any modifications to the table directly, make sure you don't have \
duplicate values in conn_id column.
    """

    @provide_session
    def check(self, session=None):
        invalid_connections = session.query(Connection.conn_id)\
            .group_by(Connection.conn_id)\
            .having(func.count() > 1)
        return (
            'Connection.conn_id={} is not unique.'.format(conn_id)
            for conn_id in invalid_connections.all()
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.configuration import conf
from airflow.upgrade.rules.base_rule import BaseRule


class FernetEnabledRule(BaseRule):
    title = "Fernet is enabled by default"

    description = (
        "The fernet mechanism is enabled by default "
        "to increase the security of the default installation."
    )

    def check(self):
        fernet_key = conf.get("core", "fernet_key")
        if not fernet_key:
            return (
                "fernet_key in airflow.cfg must be explicitly set empty as fernet mechanism is enabled"
                "by default. This means that the apache-airflow[crypto] extra-packages are always installed."
                "However, this requires that your operating system has libffi-dev installed."
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
import os


def get_rules():
    """Automatically discover all rules"""
    rule_classes = []
    path = os.path.dirname(os.path.abspath(__file__))
    for file in sorted(os.listdir(path)):
        if not file.endswith(".py") or file in ("__init__.py", "base_rule.py"):
            continue
        py_file = file[:-3]
        mod = __import__(".".join([__name__, py_file]), fromlist=[py_file])
        classes = [getattr(mod, x) for x in dir(mod) if isinstance(getattr(mod, x), type)]
        for cls in classes:
            bases = [b.__name__ for b in cls.__bases__]
            if cls.__name__ != "BaseRule" and "BaseRule" in bases:
                rule_classes.append(cls)
    # Sort rules alphabetically by class name, while maintaining that the airflow version
    # check should remain first
    return rule_classes[:1] + sorted(rule_classes[1:], key=lambda r: r.__name__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.configuration import conf
from airflow.upgrade.rules.base_rule import BaseRule


class ParsingProcessesConfigurationRule(BaseRule):
    title = "Rename max_threads to parsing_processes"

    description = "The max_threads configuration in the [scheduler] section was renamed to parsing_processes."

    def check(self):
        default = 2

        old_config_exists = conf.has_option("scheduler", "max_threads")
        new_config_exists = conf.has_option("scheduler", "parsing_processes")

        if old_config_exists and not new_config_exists:
            old_config_value = conf.get("scheduler", "max_threads")
            if old_config_value == default:
                return None
            return ["Please rename the max_threads configuration in the "
                    "[scheduler] section to parsing_processes."]
        return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.upgrade.rules.base_rule import BaseRule


class CustomExecutorsRequireFullPathRule(BaseRule):
    """
    CustomExecutorsRequireFullPathRule class to ease upgrade to Airflow 2.0
    """
    title = "Custom Executors now require full path"
    description = """\
In Airflow-2.0, loading custom executors via plugins is no longer required.
To load a custom executor, you have to provide a full path to the the custom executor module.
                  """

    def check(self):
        from airflow.plugins_manager import plugins
        executors_via_plugins = []

        for plugin in plugins:
            if plugin.executors:
                for executor in plugin.executors:
                    executors_via_plugins.append(executor.__name__)

        if executors_via_plugins:
            return (
                "Deprecation Warning: Found Custom Executor imported via a plugin."
                "From Airflow 2.0, you should use regular Python Modules to import Custom Executor."
                "You should provide a full path to the the custom executor module."
                "See the link below for more details:"
                "https://github.com/apache/airflow/blob/2.0.0/"
                "UPDATING.md#custom-executors-is-loaded-using-full-import-path \n"
                "Following Executors were imported using Plugins: \n"
                "{}".format(executors_via_plugins)
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.upgrade.rules.base_rule import BaseRule
from airflow.configuration import conf


class MesosExecutorRemovedRule(BaseRule):
    """
    MesosExecutorRemovedRule class to ease upgrade to Airflow 2.0
    """
    title = "Removal of Mesos Executor"
    description = "The Mesos Executor has been deprecated as it was not widely used and not maintained."

    def check(self):
        executor_key = conf.get(section="core", key="executor")
        if executor_key == "MesosExecutor":
            return (
                "The Mesos Executor has been deprecated as it was not widely used and not maintained."
                "Please migrate to any of the supported executors."
                "See https://airflow.apache.org/docs/stable/executor/index.html for more details."
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.configuration import conf
from airflow.upgrade.rules.base_rule import BaseRule


class HostnameCallable(BaseRule):
    title = "Unify hostname_callable option in core section"

    description = "hostname_callable option is using now only dots instead of dots and colons"

    def check(self):
        default = "socket:getfqdn"
        hostname_callable_conf = conf.get("core", "hostname_callable")
        if hostname_callable_conf == default:
            # If users use default value there's nothing they should do
            return None

        if ":" in hostname_callable_conf:
            return (
                "Error: hostname_callable `{}` "
                "contains a colon instead of a dot. please change to `{}`".format(
                    hostname_callable_conf, hostname_callable_conf.replace(":", ".")
                )
            )
        return None



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from abc import ABCMeta
from typing import List
import json

import pygments
from pygments.console import colorize
from pygments.formatters.terminal import TerminalFormatter
from pygments.lexers.markup import RstLexer

from airflow.upgrade.problem import RuleStatus
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.cli import header, get_terminal_size


class BaseFormatter(object):
    __metaclass__ = ABCMeta

    def start_checking(self, all_rules):
        # type: (List[BaseRule]) -> None
        pass

    def end_checking(self, rule_statuses):
        # type: (List[RuleStatus]) -> None

        pass

    def on_next_rule_status(self, rule_status):
        # type: (RuleStatus) -> None
        pass


class ConsoleFormatter(BaseFormatter):
    def start_checking(self, all_rules):
        print()
        header("STATUS", "=")
        print()

    def end_checking(self, rule_statuses):
        messages_count = sum(
            len(rule_status.messages)
            for rule_status in rule_statuses
        )
        if messages_count == 1:
            print("Found {} problem.".format(messages_count))
        else:
            print("Found {} problems.".format(messages_count))
        print()

        if messages_count > 0:
            header("RECOMMENDATIONS", "=")
            print()
            self.display_recommendations(rule_statuses)
        else:
            print("Not found any problems. World is beautiful. ")
            print("You can safely update Airflow to the new version.")

    @staticmethod
    def display_recommendations(rule_statuses):
        for rule_status in rule_statuses:
            # Show recommendations only if there are any messaged
            if not rule_status.messages:
                continue

            rule = rule_status.rule
            lines = [rule.title, "-" * len(rule.title)]
            if rule_status.skipped:
                lines.extend([rule_status.messages[0]])
            else:
                if rule.description:
                    lines.extend([rule.description])
                lines.extend([
                    "",
                    "Problems:",
                    "",
                ])
                lines.extend(['{:>3}.  {}'.format(i, m) for i, m in enumerate(rule_status.messages, 1)])
            msg = "\n".join(lines)

            formatted_msg = pygments.highlight(
                code=msg, formatter=TerminalFormatter(), lexer=RstLexer()
            )
            print(formatted_msg)

    def on_next_rule_status(self, rule_status):
        if rule_status.skipped:
            status = colorize("yellow", "SKIPPED")
        elif rule_status.is_success:
            status = colorize("green", "SUCCESS")
        else:
            status = colorize("red", "FAIL")
        status_line_fmt = self.prepare_status_line_format()
        print(status_line_fmt.format(rule_status.rule.title, status))

    @staticmethod
    def prepare_status_line_format():
        _, terminal_width = get_terminal_size()

        return "{:.<" + str(terminal_width - 10) + "}{:.>10}"


class JSONFormatter(BaseFormatter):
    def __init__(self, output_path):
        self.filename = output_path

    def start_checking(self, all_rules):
        print("Start looking for problems.")

    @staticmethod
    def _info_from_rule_status(rule_status):
        return {
            "rule": type(rule_status.rule).__name__,
            "title": rule_status.rule.title,
            "messages": rule_status.messages,
        }

    def end_checking(self, rule_statuses):
        formatted_results = [self._info_from_rule_status(rs) for rs in rule_statuses]
        with open(self.filename, "w+") as output_file:
            json.dump(formatted_results, output_file, indent=2)
        print("Saved result to: {}".format(self.filename))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.upgrade.rules.base_rule import BaseRule
from airflow import conf
from airflow.utils.dag_processing import list_py_file_paths


class ProperlyImportConfFromAirflow(BaseRule):
    """
    ProperlyImportConfFromAirflow class to ensure proper import of conf to work in Airflow 2.0
    """
    title = "Ensure Users Properly Import conf from Airflow"
    description = """\
    In Airflow-2.0, it's not possible to import `conf` from airflow by using `import conf from airflow`
    To ensure your code works in Airflow 2.0, you should use `from airflow.configuration import conf`.
                      """

    wrong_conf_import = "from airflow import conf"
    proper_conf_import = "from airflow.configuration import conf"

    @staticmethod
    def _conf_import_info(file_path, line_number):
        return "Affected file: {} (line {})".format(file_path, line_number)

    def _check_file(self, file_path):
        problems = []
        conf_import_check = self.wrong_conf_import
        with open(file_path, "r") as file_pointer:
            try:
                for line_number, line in enumerate(file_pointer, 1):
                    if conf_import_check in line:
                        problems.append(self._conf_import_info(file_path, line_number))
            except UnicodeDecodeError:
                problems.append("Unable to read python file {}".format(file_path))
        return problems

    def check(self):
        dag_folder = conf.get("core", "dags_folder")
        files = list_py_file_paths(directory=dag_folder, include_examples=False)
        problems = []
        for file in files:
            if not file.endswith(".py"):
                continue
            problems.extend(self._check_file(file))
        return problems



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import itertools
from typing import NamedTuple, Optional, List
import os
from cached_property import cached_property
from packaging.version import Version

from airflow import conf
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.upgrade.rules.renamed_classes import ALL
from airflow.utils.dag_processing import list_py_file_paths

try:
    from importlib_metadata import PackageNotFoundError, distribution
except ImportError:
    from importlib.metadata import PackageNotFoundError, distribution


class ImportChange(
    NamedTuple(
        "ImportChange",
        [("old_path", str), ("new_path", str), ("providers_package", Optional[str])],
    )
):
    def info(self, file_path=None):
        msg = "Using `{}` should be replaced by `{}`".format(
            self.old_path, self.new_path
        )
        if file_path:
            msg += ". Affected file: {}".format(file_path)
        return msg

    @cached_property
    def old_path_without_classname(self):
        part = self.old_path.split(".")
        part.pop()
        return ".".join(part)

    @cached_property
    def old_class(self):
        return self.old_path.split(".")[-1]

    @cached_property
    def new_class(self):
        return self.new_path.split(".")[-1]

    @classmethod
    def provider_stub_from_module(cls, module):
        if "providers" not in module:
            return None

        # [2:] strips off the airflow.providers. part
        parts = module.split(".")[2:]
        if parts[0] in ('apache', 'cncf', 'microsoft'):
            return '-'.join(parts[:2])
        return parts[0]

    @classmethod
    def from_new_old_paths(cls, new_path, old_path):
        providers_package = cls.provider_stub_from_module(new_path)
        return cls(
            old_path=old_path, new_path=new_path, providers_package=providers_package
        )


class ImportChangesRule(BaseRule):
    title = "Changes in import paths of hooks, operators, sensors and others"
    description = (
        "Many hooks, operators and other classes has been renamed and moved. Those changes were part of "
        "unifying names and imports paths as described in AIP-21.\nThe `contrib` folder has been replaced "
        "by `providers` directory and packages:\n"
        "https://github.com/apache/airflow#backport-packages"
    )

    current_airflow_version = Version(__import__("airflow").__version__)

    if current_airflow_version < Version("2.0.0"):

        def _filter_incompatible_renames(arg):
            new_path = arg[0]
            return (
                not new_path.startswith("airflow.operators")
                and not new_path.startswith("airflow.sensors")
                and not new_path.startswith("airflow.hooks")
            )

    else:
        # Everything allowed on 2.0.0+
        def _filter_incompatible_renames(arg):
            return True

    ALL_CHANGES = [
        ImportChange.from_new_old_paths(*args)
        for args in filter(_filter_incompatible_renames, ALL)
    ]  # type: List[ImportChange]

    del _filter_incompatible_renames

    @staticmethod
    def _check_file(file_path):
        problems = []
        providers = set()
        with open(file_path, "r") as file:
            try:
                content = file.read()
                for change in ImportChangesRule.ALL_CHANGES:
                    if change.old_path_without_classname in content and change.old_class in content:
                        problems.append(change.info(file_path))
                        if change.providers_package:
                            providers.add(change.providers_package)
            except UnicodeDecodeError:
                problems.append("Unable to read python file {}".format(file_path))
        return problems, providers

    @staticmethod
    def _check_missing_providers(providers):

        current_airflow_version = Version(__import__("airflow").__version__)
        if current_airflow_version >= Version("2.0.0"):
            prefix = "apache-airflow-providers-"
        else:
            prefix = "apache-airflow-backport-providers-"

        for provider in providers:
            dist_name = prefix + provider
            try:
                distribution(dist_name)
            except PackageNotFoundError:
                yield "Please install `{}`".format(dist_name)

    def check(self):
        dag_folder = conf.get("core", "dags_folder")
        files = list_py_file_paths(directory=dag_folder, include_examples=False)
        files = [file for file in files if os.path.splitext(file)[1] == ".py"]
        problems = []
        providers = set()
        # Split in to two groups - install backports first, then make changes
        for file in files:
            new_problems, new_providers = self._check_file(file)
            problems.extend(new_problems)
            providers |= new_providers

        return itertools.chain(
            self._check_missing_providers(sorted(providers)),
            problems,
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.upgrade.rules.base_rule import BaseRule
from airflow.www_rbac.utils import CustomSQLAInterface


class UseCustomSQLAInterfaceClassRule(BaseRule):
    title = "Use CustomSQLAInterface instead of SQLAInterface for custom data models."

    description = """\
From Airflow 2.0, if you want to define your own Flask App Builder models you need to
use CustomSQLAInterface instead of SQLAInterface.

For Non-RBAC replace:

`from flask_appbuilder.models.sqla.interface import SQLAInterface`
`datamodel = SQLAInterface(your_data_model)`

with RBAC (in 1.10):

`from airflow.www_rbac.utils import CustomSQLAInterface`
`datamodel = CustomSQLAInterface(your_data_model)`

and in 2.0:

`from airflow.www.utils import CustomSQLAInterface`
`datamodel = CustomSQLAInterface(your_data_model)`
                  """

    def check(self):

        from airflow.plugins_manager import flask_appbuilder_views

        plugins_with_sqlainterface_data_model_instance = []

        if flask_appbuilder_views:
            for view_obj in flask_appbuilder_views:
                if not isinstance(view_obj.get("view").datamodel, CustomSQLAInterface):
                    plugins_with_sqlainterface_data_model_instance.append(view_obj.get("name"))

        if plugins_with_sqlainterface_data_model_instance:
            return (
                "Deprecation Warning: The following views: {} have "
                "data models instantiated "
                "from the SQLAInterface class.\n".format(plugins_with_sqlainterface_data_model_instance) +
                "See: "
                "https://github.com/apache/airflow/blob/master/"
                "UPDATING.md#use-customsqlainterface-instead-of-sqlainterface-for-custom-data-models"
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.models import Connection
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.db import provide_session


class SparkJDBCOperatorConnIdRule(BaseRule):
    title = "Check Spark JDBC Operator default connection name"

    description = """\
In Airflow 1.10.x, the default value for SparkJDBCOperator class 'conn_id' is 'spark-default'.
From Airflow 2.0, it has been changed to 'spark_default' to conform with the naming conventions
of all other connection names.
    """

    @provide_session
    def check(self, session=None):
        for conn in session.query(Connection.conn_id):
            if conn.conn_id == 'spark-default':
                return (
                    "Deprecation Warning: From Airflow 2.0, the default value of 'conn_id' argument of "
                    "SparkJDBCOperator class has been changed to 'spark_default' to conform with the naming "
                    "conventions of all other connection names. Please rename the connection with "
                    "id 'spark-default' to 'spark_default' or explicitly pass 'spark-default' "
                    "to the operator. See the link below for details: "
                    "https://github.com/apache/airflow/blob/2.0.0/"
                    "UPDATING.md#sparkjdbchook-default-connection"
                )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from packaging.version import InvalidVersion, Version

from airflow.configuration import conf
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.db import provide_session


class DatabaseVersionCheckRule(BaseRule):
    title = "Check versions of PostgreSQL, MySQL, and SQLite to ease upgrade to Airflow 2.0"

    description = """\
From Airflow 2.0, the following database versions are supported:
PostgreSQl - 9.6, 10, 11, 12, 13;
MySQL - 5.7, 8;
SQLite - 3.15+
    """

    @provide_session
    def check(self, session=None):
        return self._check(session=session)

    def should_skip(self):
        try:
            self._check()
        except InvalidVersion:
            return "Unable to parse DB version, skipped!"

    @staticmethod
    @provide_session
    def _check(session=None):

        more_info = "See link below for more details: https://github.com/apache/airflow#requirements"

        conn_str = conf.get(section="core", key="sql_alchemy_conn")

        if "sqlite" in conn_str:
            min_req_sqlite_version = Version('3.15')
            installed_sqlite_version = Version(session.execute('select sqlite_version();').scalar())
            if installed_sqlite_version < min_req_sqlite_version:
                return "From Airflow 2.0, SQLite version below {} is no longer supported. \n{}".format(
                    min_req_sqlite_version, more_info
                )

        elif "postgres" in conn_str:
            min_req_postgres_version = Version('9.6')
            installed_postgres_version = Version(session.execute('SHOW server_version;').scalar())
            if installed_postgres_version < min_req_postgres_version:
                return "From Airflow 2.0, PostgreSQL version below {} is no longer supported. \n{}".format(
                    min_req_postgres_version, more_info
                )

        elif "mysql" in conn_str:
            min_req_mysql_version = Version('5.7')
            # special treatment is needed here, because MySQL version may include a suffix like '-log'
            installed_mysql_version = Version(session.execute('SELECT VERSION();').scalar().split('-')[0])
            if installed_mysql_version < min_req_mysql_version:
                return "From Airflow 2.0, MySQL version below {} is no longer supported. \n{}".format(
                    min_req_mysql_version, more_info
                )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import absolute_import

import re
import os
from airflow import conf
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.dag_processing import list_py_file_paths


class ChainBetweenDAGAndOperatorNotAllowedRule(BaseRule):

    title = "Chain between DAG and operator not allowed."

    description = "Assigning task to a DAG using bitwise shift (bit-shift) operators are no longer supported."

    def _change_info(self, file_path, line_number):
        return "{} Affected file: {} (line {})".format(
            self.title, file_path, line_number
        )

    def _check_file(self, file_path):
        problems = []
        with open(file_path, "r") as file_pointer:
            try:
                lines = file_pointer.readlines()

                python_space = r"\s*\\?\s*\n?\s*"
                # Find all the dag variable names.
                dag_vars = re.findall(r"([A-Za-z0-9_]+){}={}DAG\(".format(python_space, python_space),
                                      "".join(lines))
                history = ""
                for line_number, line in enumerate(lines, 1):
                    # Someone could have put the bitshift operator on a different line than the dag they
                    # were using it on, so search for dag >> or << dag in all previous lines that did
                    # not contain a logged issue.
                    history += line
                    matches = [
                        re.search(r"DAG\([^\)]+\){}>>".format(python_space), history),
                        re.search(r"<<{}DAG\(".format(python_space), history)
                    ]
                    for dag_var in dag_vars:
                        matches.extend([
                            re.search(r"(\s|^){}{}>>".format(dag_var, python_space), history),
                            re.search(r"<<\s*{}{}".format(python_space, dag_var), history),
                        ])
                    if any(matches):
                        problems.append(self._change_info(file_path, line_number))
                        # If we found a problem, clear our history so we don't re-log the problem
                        # on the next line.
                        history = ""
            except UnicodeDecodeError:
                problems.append("Unable to read python file {}".format(file_path))
        return problems

    def check(self):
        dag_folder = conf.get("core", "dags_folder")
        file_paths = list_py_file_paths(directory=dag_folder, include_examples=False)
        file_paths = [file for file in file_paths if os.path.splitext(file)[1] == ".py"]
        problems = []
        for file_path in file_paths:
            problems.extend(self._check_file(file_path))
        return problems



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.configuration import conf
from airflow.models.dagbag import DagBag
from airflow.upgrade.rules.base_rule import BaseRule
from airflow.utils.db import provide_session


def check_task_for_metaclasses(task):
    class_type = type(task.__class__)
    if class_type != type:
        res = (
            "Class {class_name} contained invalid custom metaclass "
            "{metaclass_name}. Custom metaclasses for operators are not "
            "allowed in Airflow 2.0. Please remove this custom metaclass.".format(
                class_name=task.__class__, metaclass_name=class_type
            )
        )
        return res
    else:
        return None


class BaseOperatorMetaclassRule(BaseRule):
    title = "Ensure users are not using custom metaclasses in custom operators"

    description = """\
In Airflow 2.0, we require that all custom operators use the BaseOperatorMeta metaclass.\
To ensure this, we can no longer allow custom metaclasses in custom operators.
    """

    @provide_session
    def check(self, session=None):
        dagbag = DagBag(dag_folder=conf.get("core", "dags_folder"), include_examples=False)
        for dag_id, dag in dagbag.dags.items():
            for task in dag.tasks:
                res = check_task_for_metaclasses(task)
                if res:
                    yield res



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from airflow.hooks.base_hook import BaseHook
from airflow.upgrade.rules.base_rule import BaseRule


def check_get_pandas_df(cls):
    try:
        cls.__new__(cls).get_pandas_df("fake SQL")
        return return_error_string(cls, "get_pandas_df")
    except NotImplementedError:
        pass
    except Exception:
        return return_error_string(cls, "get_pandas_df")


def check_run(cls):
    try:
        cls.__new__(cls).run("fake SQL")
        return return_error_string(cls, "run")
    except Exception:
        pass


def check_get_records(cls):
    try:
        cls.__new__(cls).get_records("fake SQL")
        return return_error_string(cls, "get_records")
    except Exception:
        pass


def return_error_string(cls, method):
    return (
        "Class {} incorrectly implements the function {} while inheriting from BaseHook. "
        "Please make this class inherit from airflow.hooks.db_api_hook.DbApiHook instead".format(
            cls, method
        )
    )


def get_all_non_dbapi_children():
    basehook_children = [
        child for child in BaseHook.__subclasses__() if child.__name__ != "DbApiHook"
    ]
    res = basehook_children[:]
    while basehook_children:
        next_generation = []
        for child in basehook_children:
            subclasses = child.__subclasses__()
            if subclasses:
                next_generation.extend(subclasses)
        res.extend(next_generation)
        basehook_children = next_generation
    return res


class DbApiRule(BaseRule):
    title = "Hooks that run DB functions must inherit from DBApiHook"

    description = (
        "Hooks that run DB functions must inherit from DBApiHook instead of BaseHook"
    )

    def check(self):
        basehook_subclasses = get_all_non_dbapi_children()
        incorrect_implementations = []
        for child in basehook_subclasses:
            pandas_df = check_get_pandas_df(child)
            if pandas_df:
                incorrect_implementations.append(pandas_df)
            run = check_run(child)
            if run:
                incorrect_implementations.append(run)
            get_records = check_get_records(child)
            if get_records:
                incorrect_implementations.append(get_records)
        return incorrect_implementations



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import
from typing import NamedTuple, List, Iterable

from airflow.upgrade.rules.base_rule import BaseRule


class RuleStatus(NamedTuple(
    'RuleStatus',
    [
        ('rule', BaseRule),
        ('messages', List[str]),
        ('skipped', bool)
    ]
)):
    @property
    def is_success(self):
        return len(self.messages) == 0

    @property
    def is_problem(self):
        return not self.skipped and not self.is_success

    @classmethod
    def from_rule(cls, rule):
        # type: (BaseRule) -> RuleStatus
        msg = rule.should_skip()
        if msg:
            return cls(rule=rule, messages=[msg], skipped=True)

        messages = []  # type: List[str]
        result = rule.check()
        if isinstance(result, str):
            messages = [result]
        elif isinstance(result, Iterable):
            messages = list(result)
        return cls(rule=rule, messages=messages, skipped=False)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from airflow.configuration import conf
from airflow.upgrade.rules.base_rule import BaseRule


class SendGridEmailerMovedRule(BaseRule):
    title = "SendGrid email uses old airflow.contrib module"

    description = """
The SendGrid module `airflow.contrib.utils.sendgrid` was moved to `airflow.providers.sendgrid.utils.emailer`.
    """

    def check(self):
        email_conf = conf.get(section="email", key="email_backend")
        email_contrib_path = "airflow.contrib.utils.sendgrid"
        if email_contrib_path in email_conf:
            email_provider_path = "airflow.providers.sendgrid.utils.emailer"
            msg = "Email backend option uses airflow.contrib Sendgrid module. " \
                  + "Please use new module: {}".format(email_provider_path)
            return [msg]



#                               Apache License
#                         Version 2.0, January 2004
#                      http://www.apache.org/licenses/
#
# TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION
#
# 1. Definitions.
#
#    "License" shall mean the terms and conditions for use, reproduction,
#    and distribution as defined by Sections 1 through 9 of this document.
#
#    "Licensor" shall mean the copyright owner or entity authorized by
#    the copyright owner that is granting the License.
#
#    "Legal Entity" shall mean the union of the acting entity and all
#    other entities that control, are controlled by, or are under common
#    control with that entity. For the purposes of this definition,
#    "control" means (i) the power, direct or indirect, to cause the
#    direction or management of such entity, whether by contract or
#    otherwise, or (ii) ownership of fifty percent (50%) or more of the
#    outstanding shares, or (iii) beneficial ownership of such entity.
#
#    "You" (or "Your") shall mean an individual or Legal Entity
#    exercising permissions granted by this License.
#
#    "Source" form shall mean the preferred form for making modifications,
#    including but not limited to software source code, documentation
#    source, and configuration files.
#
#    "Object" form shall mean any form resulting from mechanical
#    transformation or translation of a Source form, including but
#    not limited to compiled object code, generated documentation,
#    and conversions to other media types.
#
#    "Work" shall mean the work of authorship, whether in Source or
#    Object form, made available under the License, as indicated by a
#    copyright notice that is included in or attached to the work
#    (an example is provided in the Appendix below).
#
#    "Derivative Works" shall mean any work, whether in Source or Object
#    form, that is based on (or derived from) the Work and for which the
#    editorial revisions, annotations, elaborations, or other modifications
#    represent, as a whole, an original work of authorship. For the purposes
#    of this License, Derivative Works shall not include works that remain
#    separable from, or merely link (or bind by name) to the interfaces of,
#    the Work and Derivative Works thereof.
#
#    "Contribution" shall mean any work of authorship, including
#    the original version of the Work and any modifications or additions
#    to that Work or Derivative Works thereof, that is intentionally
#    submitted to Licensor for inclusion in the Work by the copyright owner
#    or by an individual or Legal Entity authorized to submit on behalf of
#    the copyright owner. For the purposes of this definition, "submitted"
#    means any form of electronic, verbal, or written communication sent
#    to the Licensor or its representatives, including but not limited to
#    communication on electronic mailing lists, source code control systems,
#    and issue tracking systems that are managed by, or on behalf of, the
#    Licensor for the purpose of discussing and improving the Work, but
#    excluding communication that is conspicuously marked or otherwise
#    designated in writing by the copyright owner as "Not a Contribution."
#
#    "Contributor" shall mean Licensor and any individual or Legal Entity
#    on behalf of whom a Contribution has been received by Licensor and
#    subsequently incorporated within the Work.
#
# 2. Grant of Copyright License. Subject to the terms and conditions of
#    this License, each Contributor hereby grants to You a perpetual,
#    worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#    copyright license to reproduce, prepare Derivative Works of,
#    publicly display, publicly perform, sublicense, and distribute the
#    Work and such Derivative Works in Source or Object form.
#
# 3. Grant of Patent License. Subject to the terms and conditions of
#    this License, each Contributor hereby grants to You a perpetual,
#    worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#    (except as stated in this section) patent license to make, have made,
#    use, offer to sell, sell, import, and otherwise transfer the Work,
#    where such license applies only to those patent claims licensable
#    by such Contributor that are necessarily infringed by their
#    Contribution(s) alone or by combination of their Contribution(s)
#    with the Work to which such Contribution(s) was submitted. If You
#    institute patent litigation against any entity (including a
#    cross-claim or counterclaim in a lawsuit) alleging that the Work
#    or a Contribution incorporated within the Work constitutes direct
#    or contributory patent infringement, then any patent licenses
#    granted to You under this License for that Work shall terminate
#    as of the date such litigation is filed.
#
# 4. Redistribution. You may reproduce and distribute copies of the
#    Work or Derivative Works thereof in any medium, with or without
#    modifications, and in Source or Object form, provided that You
#    meet the following conditions:
#
#    (a) You must give any other recipients of the Work or
#        Derivative Works a copy of this License; and
#
#    (b) You must cause any modified files to carry prominent notices
#        stating that You changed the files; and
#
#    (c) You must retain, in the Source form of any Derivative Works
#        that You distribute, all copyright, patent, trademark, and
#        attribution notices from the Source form of the Work,
#        excluding those notices that do not pertain to any part of
#        the Derivative Works; and
#
#    (d) If the Work includes a "NOTICE" text file as part of its
#        distribution, then any Derivative Works that You distribute must
#        include a readable copy of the attribution notices contained
#        within such NOTICE file, excluding those notices that do not
#        pertain to any part of the Derivative Works, in at least one
#        of the following places: within a NOTICE text file distributed
#        as part of the Derivative Works; within the Source form or
#        documentation, if provided along with the Derivative Works; or,
#        within a display generated by the Derivative Works, if and
#        wherever such third-party notices normally appear. The contents
#        of the NOTICE file are for informational purposes only and
#        do not modify the License. You may add Your own attribution
#        notices within Derivative Works that You distribute, alongside
#        or as an addendum to the NOTICE text from the Work, provided
#        that such additional attribution notices cannot be construed
#        as modifying the License.
#
#    You may add Your own copyright statement to Your modifications and
#    may provide additional or different license terms and conditions
#    for use, reproduction, or distribution of Your modifications, or
#    for any such Derivative Works as a whole, provided Your use,
#    reproduction, and distribution of the Work otherwise complies with
#    the conditions stated in this License.
#
# 5. Submission of Contributions. Unless You explicitly state otherwise,
#    any Contribution intentionally submitted for inclusion in the Work
#    by You to the Licensor shall be under the terms and conditions of
#    this License, without any additional terms or conditions.
#    Notwithstanding the above, nothing herein shall supersede or modify
#    the terms of any separate license agreement you may have executed
#    with Licensor regarding such Contributions.
#
# 6. Trademarks. This License does not grant permission to use the trade
#    names, trademarks, service marks, or product names of the Licensor,
#    except as required for reasonable and customary use in describing the
#    origin of the Work and reproducing the content of the NOTICE file.
#
# 7. Disclaimer of Warranty. Unless required by applicable law or
#    agreed to in writing, Licensor provides the Work (and each
#    Contributor provides its Contributions) on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#    implied, including, without limitation, any warranties or conditions
#    of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#    PARTICULAR PURPOSE. You are solely responsible for determining the
#    appropriateness of using or redistributing the Work and assume any
#    risks associated with Your exercise of permissions under this License.
#
# 8. Limitation of Liability. In no event and under no legal theory,
#    whether in tort (including negligence), contract, or otherwise,
#    unless required by applicable law (such as deliberate and grossly
#    negligent acts) or agreed to in writing, shall any Contributor be
#    liable to You for damages, including any direct, indirect, special,
#    incidental, or consequential damages of any character arising as a
#    result of this License or out of the use or inability to use the
#    Work (including but not limited to damages for loss of goodwill,
#    work stoppage, computer failure or malfunction, or any and all
#    other commercial damages or losses), even if such Contributor
#    has been advised of the possibility of such damages.
#
# 9. Accepting Warranty or Additional Liability. While redistributing
#    the Work or Derivative Works thereof, You may choose to offer,
#    and charge a fee for, acceptance of support, warranty, indemnity,
#    or other liability obligations and/or rights consistent with this
#    License. However, in accepting such obligations, You may act only
#    on Your own behalf and on Your sole responsibility, not on behalf
#    of any other Contributor, and only if You agree to indemnify,
#    defend, and hold each Contributor harmless for any liability
#    incurred by, or claims asserted against, such Contributor by reason
#    of your accepting any such warranty or additional liability.
#
# END OF TERMS AND CONDITIONS
#
# APPENDIX: How to apply the Apache License to your work.
#
#    To apply the Apache License to your work, attach the following
#    boilerplate notice, with the fields enclosed by brackets "[]"
#    replaced with your own identifying information. (Don't include
#    the brackets!)  The text should be enclosed in the appropriate
#    comment syntax for the file format. We also recommend that a
#    file or class name and description of purpose be included on the
#    same "printed page" as the copyright notice for easier
#    identification within third-party archives.
#
# Copyright [yyyy] [name of copyright owner]
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ============================================================================
#    APACHE AIRFLOW SUBCOMPONENTS:
#
#    The Apache Airflow project contains subcomponents with separate copyright
#    notices and license terms. Your use of the source code for the these
#    subcomponents is subject to the terms and conditions of the following
#    licenses.
#
#
# ========================================================================
# Third party Apache 2.0 licenses
# ========================================================================
#
# The following components are provided under the Apache 2.0 License.
# See project link for details. The text of each license is also included
# at licenses/LICENSE-[project].txt.
#
#     (ALv2 License) hue v4.3.0 (https://github.com/cloudera/hue/)
#     (ALv2 License) jqclock v2.3.0 (https://github.com/JohnRDOrazio/jQuery-Clock-Plugin)
#     (ALv2 License) bootstrap3-typeahead v4.0.2 (https://github.com/bassjobsen/Bootstrap-3-Typeahead)
#     (ALv2 License) airflow.contrib.auth.backends.github_enterprise_auth
#
# ========================================================================
# MIT licenses
# ========================================================================
#
# The following components are provided under the MIT License. See project link for details.
# The text of each license is also included at licenses/LICENSE-[project].txt.
#
#     (MIT License) jquery v2.1.4 (https://jquery.org/license/)
#     (MIT License) dagre-d3 v0.6.1 (https://github.com/cpettitt/dagre-d3)
#     (MIT License) bootstrap v3.2 (https://github.com/twbs/bootstrap/)
#     (MIT License) d3-tip v0.6.3 (https://github.com/Caged/d3-tip)
#     (MIT License) dataTables v1.10.10 (https://datatables.net)
#     (MIT License) WebGL-2D (git-commit 9a7ec26) (https://github.com/gameclosure/webgl-2d)
#     (MIT License) Underscorejs v1.5.0 (http://underscorejs.org)
#     (MIT License) Bootstrap Toggle v2.2.0 (http://www.bootstraptoggle.com)
#     (MIT License) normalize.css v3.0.2 (http://necolas.github.io/normalize.css/)
#     (MIT License) ElasticMock v1.3.2 (https://github.com/vrcmarcos/elasticmock)
#     (MIT License) MomentJS v2.22.2 (http://momentjs.com/)
#     (MIT License) moment-strftime v0.5.0 (https://github.com/benjaminoakes/moment-strftime)
#     (MIT License) python-slugify v4.0.0 (https://github.com/un33k/python-slugify)
#     (MIT License) python-nvd3 v0.15.0 (https://github.com/areski/python-nvd3)
#     (MIT License) eonasdan-bootstrap-datetimepicker v4.17.37 (https://github.com/eonasdan/bootstrap-datetimepicker/)
#
# ========================================================================
# BSD 2-Clause licenses
# ========================================================================
# The following components are provided under the BSD 2-Clause license.
# See file headers and project links for details.
# The text of each license is also included at licenses/LICENSE-[project].txt.
#
#     (BSD 2 License) flask-kerberos v1.0.4 (https://github.com/mkomitee/flask-kerberos)
#
# ========================================================================
# BSD 3-Clause licenses
# ========================================================================
# The following components are provided under the BSD 3-Clause license. See project links for details.
# The text of each license is also included at licenses/LICENSE-[project].txt.
#
#     (BSD 3 License) Ace v1.1.8 (https://github.com/ajaxorg/ace)
#     (BSD 3 License) d3js v3.5.17 (https://d3js.org)
#     (BSD 3 License) parallel-coordinates v0.7.0 (http://syntagmatic.github.com/parallel-coordinates/)
#     (BSD 3 License) scikit-learn v0.19.1 (https://github.com/scikit-learn/scikit-learn)

__import__("setuptools").setup()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

import os
import re
import logging
import warnings
from airflow.utils.dag_processing import correct_maybe_zipped, list_py_file_paths
from airflow import conf
from airflow.models.dagbag import DagBag
from airflow.upgrade.rules.base_rule import BaseRule


class NoAdditionalArgsInOperatorsRule(BaseRule):
    title = "No additional argument allowed in BaseOperator."

    description = """\
Passing unrecognized arguments to operators is not allowed in Airflow 2.0 anymore,
and will cause an exception.
                  """

    def check(self, dags_folder=None):
        if not dags_folder:
            dags_folder = conf.get("core", "dags_folder")

        logger = logging.root
        old_level = logger.level
        try:
            logger.setLevel(logging.ERROR)
            dagbag = DagBag(dag_folder=os.devnull, include_examples=False, store_serialized_dags=False)
            dags_folder = correct_maybe_zipped(dags_folder)

            # Each file in the DAG folder is parsed individually
            for filepath in list_py_file_paths(dags_folder,
                                               safe_mode=conf.getboolean('core', 'DAG_DISCOVERY_SAFE_MODE'),
                                               include_examples=False):
                try:
                    with warnings.catch_warnings(record=True) as captured_warnings:
                        _ = dagbag.process_file(
                            filepath,
                            safe_mode=conf.getboolean('core', 'DAG_DISCOVERY_SAFE_MODE'))
                except Exception:
                    pass

                for warning in captured_warnings:
                    if warning.category in (DeprecationWarning, PendingDeprecationWarning) \
                            and str(warning.message).startswith("Invalid arguments were passed"):
                        m = re.match(r'''
                                    .* \(task_id:\ ([^\)]+)\) .* \n
                                    \*args:\ (.*) \n
                                     \*\*kwargs:\ (.*)
                                    ''', str(warning.message), re.VERBOSE)

                        yield "DAG file `{}` with task_id `{}` has unrecognized positional args `{}`" \
                              "and keyword args " \
                              "`{}`".format(filepath, m.group(1), m.group(2), m.group(3))
        finally:
            logger.setLevel(old_level)
